from pathlib import Path
import re
from typing import Dict, List, Mapping, Union

from gartersnake.pip import installed_packages


def read_requirements() -> List[str]:
    requirements_txt_filename = Path('requirements.txt')
    packages = []
    if requirements_txt_filename.exists():
        packages.extend(requirements_txt_filename.open().readlines())
    return packages


def missing_requirements(
    requirements: Union[List[str], Dict[str, List[str]]] = None
) -> Union[List[str], Dict[str, List[str]]]:
    if requirements is None:
        requirements = read_requirements()

    if isinstance(requirements, Mapping):
        missing_packages = missing_requirements(list(requirements))
        missing_dependencies = {}
        for requirement, subrequirements in requirements.items():
            missing_subpackages = missing_requirements(subrequirements)
            if requirement in missing_packages or len(missing_subpackages) > 0:
                missing_dependencies[requirement] = missing_subpackages
        return missing_dependencies
    else:
        return [
            required_package
            for required_package in requirements
            if re.split('<|<=|==|>=|>', required_package)[0].lower()
            not in installed_packages()
        ]



import subprocess
import sys
from typing import List

try:
    from importlib import metadata as importlib_metadata
except ImportError:  # for Python<3.8
    subprocess.run(
        f'{sys.executable} -m pip install importlib_metadata',
        shell=True,
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
    )
    import importlib_metadata


def installed_packages() -> List[str]:
    installed_distributions = importlib_metadata.distributions()
    return [
        distribution.metadata['Name'].lower()
        for distribution in installed_distributions
        if distribution.metadata['Name'] is not None
    ]



from pathlib import Path
import subprocess
import sys
from typing import List

from gartersnake.missing import missing_requirements


def is_conda() -> bool:
    return (Path(sys.prefix) / 'conda-meta').exists()


def install_conda_requirements(requirements: List[str] = None, overwrite: bool = False):
    if is_conda():
        print(f'found conda environment at {sys.prefix}')
    else:
        raise EnvironmentError('conda environment not detected')

    if overwrite:
        missing_packages = requirements
    else:
        missing_packages = missing_requirements(requirements)

    conda_packages = []
    try:
        subprocess.check_output(
            f'conda install -y {" ".join(missing_packages)}',
            shell=True,
            stderr=subprocess.STDOUT,
        )
    except subprocess.CalledProcessError as error:
        output = error.output.decode()
        package_not_found_start = 'PackagesNotFoundError: The following packages are not available from current channels:\n\n'
        package_not_found_stop = '\n\nCurrent channels:'
        if package_not_found_start in output:
            non_conda_packages = [
                package.strip()
                for package in output[
                    output.index(package_not_found_start) : output.index(
                        package_not_found_stop
                    )
                ].splitlines()[2:]
            ]
            conda_packages = [
                package
                for package in missing_packages
                if not any(
                    non_conda_package in package for non_conda_package in non_conda_packages
                )
            ]

            print(
                f'found {len(conda_packages)} conda packages (out of {len(missing_packages)}) - {conda_packages}'
            )

    try:
        subprocess.run(
            f'conda install -y {" ".join(conda_packages)}',
            shell=True,
            stderr=subprocess.DEVNULL,
        )
    except subprocess.CalledProcessError:
        print('batch install failed; attempting individual installs...')
        for dependency in conda_packages:
            try:
                subprocess.run(
                    f'conda install -y {dependency}', shell=True, stderr=subprocess.DEVNULL,
                )
            except subprocess.CalledProcessError:
                continue



from gartersnake.conda import install_conda_requirements, is_conda
from gartersnake.missing import missing_requirements
from gartersnake.pip import installed_packages
from gartersnake.version import vcs_version
from gartersnake.windows import install_windows_requirements, is_windows



from gartersnake.pip import installed_packages


def vcs_version() -> str:
    try:
        if 'dunamai' not in installed_packages():
            print('`dunamai` is not installed; version will default to `0.0.0`')

        from dunamai import Version

        version = Version.from_any_vcs().serialize()
    except (ModuleNotFoundError, RuntimeError) as error:
        print(error)
        version = '0.0.0'

    return version



import os
import subprocess
import sys
from typing import Dict, List, Mapping, Union

from gartersnake.missing import missing_requirements, read_requirements
from gartersnake.pip import installed_packages


def is_windows() -> bool:
    return os.name == 'nt'


def install_windows_requirements(
    requirements: Union[List[str], Dict[str, List[str]]] = None, overwrite: bool = False
):
    if not is_windows():
        raise EnvironmentError('Windows environment not detected')

    if requirements is None:
        requirements = read_requirements()

    if not isinstance(requirements, Mapping):
        requirements = {requirement: [] for requirement in requirements}

    if overwrite:
        missing_packages = requirements
    else:
        missing_packages = missing_requirements(requirements)

    print(f'attempting to install {len(missing_packages)} packages with `pipwin`')

    if 'pipwin' not in installed_packages():
        raise ImportError(
            'Windows detected but `pipwin` is not installed. Please install `pipwin` to continue.'
        )

    subprocess.run(f'{sys.executable} -m pipwin refresh', shell=True)

    for requirement, subrequirements in missing_packages.items():
        failed_pipwin_packages = []

        # since we don't know the dependencies here, repeat this process n number of times
        # (worst case is `O(n)`, where the first package is dependent on all the others)
        for _ in range(1 + len(subrequirements)):
            for package_name in subrequirements + [requirement]:
                if requirement in missing_requirements(
                    requirements
                ) or package_name in missing_requirements(subrequirements):
                    try:
                        subprocess.run(
                            f'{sys.executable} -m pip install {package_name.lower()}',
                            check=True,
                            shell=True,
                            stderr=subprocess.DEVNULL,
                        )
                        if package_name in failed_pipwin_packages:
                            failed_pipwin_packages.remove(package_name)
                    except subprocess.CalledProcessError:
                        try:
                            subprocess.run(
                                f'{sys.executable} -m pipwin install {package_name.lower()}',
                                check=True,
                                shell=True,
                                stderr=subprocess.DEVNULL,
                            )
                        except subprocess.CalledProcessError:
                            failed_pipwin_packages.append(package_name)

            if len(failed_pipwin_packages) == 0:
                break



import subprocess
import sys
from typing import List

from setuptools import config, find_packages, setup

try:
    from importlib import metadata as importlib_metadata
except ImportError:  # for Python<3.8
    subprocess.run(
        f'{sys.executable} -m pip install importlib_metadata',
        shell=True,
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
    )
    import importlib_metadata


def installed_packages() -> List[str]:
    installed_distributions = importlib_metadata.distributions()
    return [
        distribution.metadata['Name'].lower()
        for distribution in installed_distributions
        if distribution.metadata['Name'] is not None
    ]


try:
    if 'dunamai' not in installed_packages():
        subprocess.run(
            f'{sys.executable} -m pip install dunamai',
            shell=True,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )

    from dunamai import Version

    __version__ = Version.from_any_vcs().serialize()
except (ModuleNotFoundError, RuntimeError) as error:
    print(error)
    __version__ = '0.0.0'

print(f'using version {__version__}')

metadata = config.read_configuration('setup.cfg')['metadata']

setup(
    **metadata,
    version=__version__,
    long_description_content_type='text/markdown',
    packages=find_packages(),
    python_requires='>=3.6',
    setup_requires=['dunamai', 'setuptools>=41.2'],
    install_requires=[],
    extras_require={
        'testing': ['pytest', 'pytest-cov', 'pytest-xdist', 'wget'],
        'development': ['flake8', 'isort', 'oitnb'],
    },
)
